/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://eu-egee.org/partners/ for details on the copyright holders.
 * For license conditions see the license file or http://eu-egee.org/license.html
 *
 * Authors:
 *     Gabor Gombas <Gabor.Gombas@cern.ch>
 *     John Walk <J.A.Walk@rl.ac.uk>
 */

/* This file provides a layer of indirection between the public interface
   defined in ServiceDiscovery.h and the actual implementation (for example
   the R-GMA implementation in ServiceDiscoveryIRGMA.c, and the common functions
   in ServiceDiscoveryICommon.c). This allows alternative implementations to
   share any common functionality. The functions are documented in the header
   file. */

#define _GNU_SOURCE                 /* for strcasestr() */
#include <stdio.h>                  /* for NULL */
#include <stdlib.h>                 /* for free() */
#include <string.h>                 /* for strcmp() and memmove() */
#include <dlfcn.h>                  /* for dlopen() and dlsym() */
#include <errno.h>

#include "ServiceDiscovery.h"       /* Public Service Discovery interface. */
#include "ServiceDiscoveryI.h"      /* Private interface to implementation. */

#define PRIVATE static
#define PUBLIC /* empty */

#define SD_DEFAULT_PLUGIN   "file"

#define ERROR_OUTOFMEMORY  "Insufficient memory to complete API call."
#define ERROR_NOPLUGIN     "Service discovery plugin was not found."
#define ERROR_NULLPLUGIN   "Plugin cannot be NULL."
#define ERROR_NOPLUGINNAME "Plugin name is missing."
#define ERROR_PLUGINEXISTS "Plugin is already registered."
#define ERROR_BADPLUGIN    "The plugin does not implement a mandatory method."

typedef struct
{
	int			numPlugins;
	const SDPlugin		**plugins;
} SDPluginList;

PRIVATE const SDPlugin *getPluginByName(const char *pluginName);
PRIVATE const SDPlugin *nextPlugin(const SDPlugin *plugin);
PRIVATE int addPluginToList(SDPluginList *list, const SDPlugin *plugin);
PRIVATE void delPluginFromList(SDPluginList *list, const SDPlugin *plugin);
PRIVATE SDVOList *checkVOEnv(void);
PRIVATE char *dupstr(const char *);

/* Set the owner field. Use a temp. variable to avoid lvalue cast. The const
 * cast is OK since the 'const' in the API is only there to discourage
 * applications from playing with the field's value */
#define SET_OWNER(x, owner) \
	do { \
		const void **__tmp = (const void **)&(x)->_owner; \
		*__tmp = owner; \
	} while (0)


/* GLOBAL VARIABLES ***********************************************************/

/* List of registered plugins */
PRIVATE SDPluginList REGISTERED_PLUGINS;

/* List of useable plugins */
PRIVATE SDPluginList LOADED_PLUGINS;


/* PUBLIC FUNCTIONS ***********************************************************/

/* The functions that access the information system are handed off to a
   particular implementation (e.g. ServiceDiscoveryIRGMA.c). Housekeeping
   functions are implemented in ServiceDiscoveryICommon.c. */

PUBLIC SDService *SD_getService(const char *serviceName, SDException *exception)
{
	SDServiceDetails *serviceDetails = NULL;
	const SDPlugin *plugin = NULL;
	SDService *service = NULL;

	/* Make sure the first SD_freeException() does not do anything stupid */
	SD_I_setException(exception, SDStatus_SUCCESS, NULL);

	while ((plugin = nextPlugin(plugin)))
	{
		/* Free the exception set by the previous iteration */
		SD_freeException(exception);

		/* Call the plugin's getService method if there is one,
		 * or emulate it if there isn't */
		if (plugin->getService)
		{
			service = plugin->getService(serviceName, exception);
			if (service)
			{
				SET_OWNER(service, plugin);
				break;
			}
		}
		else
		{
			serviceDetails = plugin->getServiceDetails(serviceName, exception);
			if (serviceDetails)
			{
				SET_OWNER(serviceDetails, plugin);
				break;
			}
		}
	}

	if (!plugin && exception && exception->status == SDStatus_SUCCESS)
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_NOPLUGIN);

	if (serviceDetails)
	{
		/* If we are emulating the getService method, we have to make
		 * a copy of the data */
		service = (SDService *) calloc(1, sizeof(SDService));
		if (!service)
			SD_I_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY);
		else
		{
			/* Copy the basic service data */
			service->name = dupstr(serviceDetails->name);
			service->type = dupstr(serviceDetails->type);
			service->endpoint = dupstr(serviceDetails->endpoint);
			service->version = dupstr(serviceDetails->version);

			/* Check for failed dupstr()s */
			if (!service->name || !service->type || !service->endpoint ||
					!service->version)
			{
				SD_I_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY);
				SD_freeService(service);
				service = NULL;
			}
		}

		SD_freeServiceDetails(serviceDetails);
	}

	return service;
}

PUBLIC SDServiceDetails *SD_getServiceDetails(const char *serviceName, SDException *exception)
{
	SDServiceDetails *serviceDetails = NULL;
	const SDPlugin *plugin = NULL;

	/* Make sure the first SD_freeException() does not do anything stupid */
	SD_I_setException(exception, SDStatus_SUCCESS, NULL);

	while ((plugin = nextPlugin(plugin)))
	{
		/* Free the exception set by the previous iteration */
		SD_freeException(exception);

		serviceDetails = plugin->getServiceDetails(serviceName, exception);
		if (serviceDetails)
		{
			SET_OWNER(serviceDetails, plugin);
			break;
		}
	}

	if (!plugin && exception && exception->status == SDStatus_SUCCESS)
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_NOPLUGIN);

	return serviceDetails;
}

PUBLIC SDServiceDataList *SD_getServiceData(const char *serviceName, SDException *exception)
{
	SDServiceDetails *serviceDetails = NULL;
	SDServiceDataList *list = NULL;
	const SDPlugin *plugin = NULL;

	/* Make sure the first SD_freeException() does not do anything stupid */
	SD_I_setException(exception, SDStatus_SUCCESS, NULL);

	while ((plugin = nextPlugin(plugin)))
	{
		/* Free the exception set by the previous iteration */
		SD_freeException(exception);

		/* Call the plugin's getServiceData method if there is one,
		 * or emulate it if there isn't */
		if (plugin->getServiceData)
		{
			list = plugin->getServiceData(serviceName, exception);
			if (list)
			{
				SET_OWNER(list, plugin);
				break;
			}
		}
		else
		{
			serviceDetails = plugin->getServiceDetails(serviceName, exception);
			if (serviceDetails)
			{
				SET_OWNER(serviceDetails, plugin);
				break;
			}
		}
	}

	if (!plugin && exception && exception->status == SDStatus_SUCCESS)
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_NOPLUGIN);

	if (serviceDetails)
	{
		/* If we are emulating the getServiceData method, we have to make
		 * a copy of the data list */
		int i;

		list = (SDServiceDataList *) malloc(sizeof(SDServiceDataList));
		if (!list)
			SD_I_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY);
		else
		{
			SET_OWNER(list, NULL);
			list->numItems = serviceDetails->data->numItems;
			list->items = (SDServiceData *) calloc(list->numItems, sizeof(SDServiceData));
			if (!list->items)
			{
				SD_I_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY);
				SD_I_freeServiceDataList(list);
				list = NULL;
			}
		}

		for (i = 0; list && i < list->numItems; i++)
		{
			list->items[i].key = dupstr(serviceDetails->data->items[i].key);
			list->items[i].value = dupstr(serviceDetails->data->items[i].value);
			if (!list->items[i].key || !list->items[i].value)
			{
				SD_I_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY);
				SD_I_freeServiceDataList(list);
				list = NULL;
				break;
			}
		}

		SD_freeServiceDetails(serviceDetails);
	}
	return list;
}

PUBLIC char *SD_getServiceDataItem(const char *serviceName, const char *key, SDException *exception)
{
	SDServiceDetails *serviceDetails = NULL;
	SDServiceDataList *list = NULL;
	const SDPlugin *plugin = NULL;
	char *value = NULL;

	/* Make sure the first SD_freeException() does not do anything stupid */
	SD_I_setException(exception, SDStatus_SUCCESS, NULL);

	while ((plugin = nextPlugin(plugin)))
	{
		/* Free the exception set by the previous iteration */
		SD_freeException(exception);

		/* Call the plugin's getServiceDataItem method if there is one,
		 * or emulate it if there isn't */
		if (plugin->getServiceDataItem)
		{
			value = plugin->getServiceDataItem(serviceName, key, exception);
			if (value)
				break;
		}
		else if (plugin->getServiceData)
		{
			list = plugin->getServiceData(serviceName, exception);
			if (list)
			{
				SET_OWNER(list, plugin);
				break;
			}
		}
		else
		{
			serviceDetails = plugin->getServiceDetails(serviceName, exception);
			if (serviceDetails)
			{
				SET_OWNER(serviceDetails, plugin);
				break;
			}
		}
	}

	if (!plugin && exception && exception->status == SDStatus_SUCCESS)
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_NOPLUGIN);

	if (serviceDetails)
		list = serviceDetails->data;
	if (list)
	{
		/* If we are emulating the getServiceDataItem method, we have to
		 * look up the key and make a copy of the value */
		int i;

		value = NULL;
		for (i = 0; i < list->numItems; i++)
		{
			if (!strcmp(list->items[i].key, key))
			{
				value = dupstr(list->items[i].value);
				if (!value)
					SD_I_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY);
				break;
			}
		}
	}

	if (serviceDetails)
		SD_freeServiceDetails(serviceDetails);
	else if (list)
		SD_freeServiceDataList(list);

	return value;
}

PUBLIC char *SD_getServiceSite(const char *serviceName, SDException *exception)
{
	SDServiceDetails *serviceDetails = NULL;
	const SDPlugin *plugin = NULL;
	char *site = NULL;

	/* Make sure the first SD_freeException() does not do anything stupid */
	SD_I_setException(exception, SDStatus_SUCCESS, NULL);

	while ((plugin = nextPlugin(plugin)))
	{
		/* Free the exception set by the previous iteration */
		SD_freeException(exception);

		/* Call the plugin's getServiceSite method if there is one,
		 * or emulate it if there isn't */
		if (plugin->getServiceSite)
		{
			site = plugin->getServiceSite(serviceName, exception);
			if (site)
				break;
		}
		else
		{
			serviceDetails = plugin->getServiceDetails(serviceName, exception);
			if (serviceDetails)
			{
				SET_OWNER(serviceDetails, plugin);
				break;
			}
		}
	}

	if (!plugin && exception && exception->status == SDStatus_SUCCESS)
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_NOPLUGIN);

	if (serviceDetails)
	{
		/* If we are emulating the getServiceSite method, we have to
		 * make a copy of the value */
		if (serviceDetails->site)
		{
			site = dupstr(serviceDetails->site);
			if (!site)
				SD_I_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY);
		}
		else
			site = NULL;

		SD_freeServiceDetails(serviceDetails);
	}

	return site;
}

PUBLIC char *SD_getServiceWSDL(const char *serviceName, SDException *exception)
{
	SDServiceDetails *serviceDetails = NULL;
	const SDPlugin *plugin = NULL;
	char *wsdl = NULL;

	/* Make sure the first SD_freeException() does not do anything stupid */
	SD_I_setException(exception, SDStatus_SUCCESS, NULL);

	while ((plugin = nextPlugin(plugin)))
	{
		/* Free the exception set by the previous iteration */
		SD_freeException(exception);

		/* Call the plugin's getServiceWSDL method if there is one,
		 * or emulate it if there isn't */
		if (plugin->getServiceWSDL)
		{
			wsdl = plugin->getServiceWSDL(serviceName, exception);
			if (wsdl)
				break;
		}
		else
		{
			serviceDetails = plugin->getServiceDetails(serviceName, exception);
			if (serviceDetails)
			{
				SET_OWNER(serviceDetails, plugin);
				break;
			}
		}
	}

	if (!plugin && exception && exception->status == SDStatus_SUCCESS)
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_NOPLUGIN);

	if (serviceDetails)
	{
		/* If we are emulating the getServiceWSDL method, we have to
		 * make a copy of the value */
		if (serviceDetails->wsdl)
		{
			wsdl = dupstr(serviceDetails->wsdl);
			if (!wsdl)
				SD_I_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY);
		}
		else
			wsdl = NULL;

		SD_freeServiceDetails(serviceDetails);
	}

	return wsdl;
}

PUBLIC SDServiceList *SD_listAssociatedServices(const char *serviceName, const char *type, const char *site, const SDVOList *vos, SDException *exception)
{
	const SDPlugin *plugin = NULL;
	SDServiceList *result = NULL;
	SDVOList *tmpvos;

	/* Make sure the first SD_freeException() does not do anything stupid */
	SD_I_setException(exception, SDStatus_SUCCESS, NULL);

	if (!site)
    {
		site = getenv(SDEnv_SITE);
        /* treat empty string as if it does not exist */
        if (site && (strlen(site) == 0)) site = NULL;
    }
	tmpvos = NULL;
	/* Let the user override the VO if the caller did not specify
	 * one explicitely */
	if (!vos || !vos->numNames)
		tmpvos = checkVOEnv();

	while ((plugin = nextPlugin(plugin)))
	{
		/* Free the exception set by the previous iteration */
		SD_freeException(exception);

		result = plugin->listAssociatedServices(serviceName, type, site,
			tmpvos ? tmpvos : vos,
			exception);
		if (result)
		{
			SET_OWNER(result, plugin);
			break;
		}
	}

	if (!plugin && exception && exception->status == SDStatus_SUCCESS)
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_NOPLUGIN);

	if (tmpvos)
		SD_I_freeVOList(tmpvos);
	return result;
}

PUBLIC SDServiceList *SD_listServices(const char *type, const char *site, const SDVOList *vos, SDException *exception)
{
	const SDPlugin *plugin = NULL;
	SDServiceList *result = NULL;
	SDVOList *tmpvos;

	/* Make sure the first SD_freeException() does not do anything stupid */
	SD_I_setException(exception, SDStatus_SUCCESS, NULL);

	if (!site)
    {
		site = getenv(SDEnv_SITE);
        /* treat empty string as if it does not exist */
        if (site && (strlen(site) == 0)) site = NULL;
    }
	tmpvos = NULL;
	/* Let the user override the VO if the caller did not specify
	 * one explicitely */
	if (!vos || !vos->numNames)
		tmpvos = checkVOEnv();

	while ((plugin = nextPlugin(plugin)))
	{
		/* Free the exception set by the previous iteration */
		SD_freeException(exception);

		if (plugin->listServices)
		{
			result = plugin->listServices(type, site,
				tmpvos ? tmpvos : vos,
				exception);
		}
		else
		{
			result = plugin->listServicesByData(NULL, type, site,
				tmpvos ? tmpvos : vos,
				exception);
		}
		if (result)
		{
			SET_OWNER(result, plugin);
			break;
		}
	}

	if (!plugin && exception && exception->status == SDStatus_SUCCESS)
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_NOPLUGIN);

	if (tmpvos)
		SD_I_freeVOList(tmpvos);
	return result;
}

typedef int (*SD_matchEndpointHost)(const char *endpoint, const char *host);

// matching '^host$'
PRIVATE int SD_matchEndpointHost_0(const char *endpoint, const char *host) {
    return ! strcasecmp(endpoint, host);
}

// matching '^host:'
PRIVATE int SD_matchEndpointHost_1(const char *endpoint, const char *host) {
    return (! strncasecmp(endpoint, host, strlen(host)))
        && (strlen(endpoint) > strlen(host))
        && (endpoint[strlen(host)] == ':');
}

// matching '://host/'
PRIVATE int SD_matchEndpointHost_2(const char *endpoint, const char *host) {
    const char *hostp = strcasestr(endpoint, host);

    if (NULL == hostp) return 0; // host not found
    
    int index = hostp - endpoint;
    if (index < 3) return 0;    //  cannot match '://'
    if (strlen(endpoint) < index + strlen(host) + 1) return 0; // last slash
    return (hostp[-3] == ':') && (hostp[-2] == '/') && (hostp[-1] == '/')
        && (hostp[strlen(host)] == '/');
}

// matching '://host:'
PRIVATE int SD_matchEndpointHost_3(const char *endpoint, const char *host) {
    const char *hostp = strcasestr(endpoint, host);

    if (NULL == hostp) return 0; // host not found
    
    int index = hostp - endpoint;
    if (index < 3) return 0;    //  cannot match '://'
    if (strlen(endpoint) < index + strlen(host) + 1) return 0; // last slash
    return (hostp[-3] == ':') && (hostp[-2] == '/') && (hostp[-1] == '/')
        && (hostp[strlen(host)] == ':');
}

// matching 'host' anywhere
PRIVATE int SD_matchEndpointHost_4(const char *endpoint, const char *host) {
    return strcasestr(endpoint, host) != NULL;
}

static SD_matchEndpointHost SD_matchEndpointHostFunctions[] = {
    &SD_matchEndpointHost_0,
    &SD_matchEndpointHost_1,
    &SD_matchEndpointHost_2,
    &SD_matchEndpointHost_3,
    &SD_matchEndpointHost_4
};
static int SD_num_matches = sizeof(SD_matchEndpointHostFunctions) / sizeof(SD_matchEndpointHost);

PUBLIC SDServiceList *SD_listServicesByHost(const char *type, const char *host, const SDVOList *vos, SDException *exception)
{
	const SDPlugin *plugin = NULL;
	SDServiceList *result = NULL;
	SDVOList *tmpvos;
	int i,j;
        int numServices;

	/* Make sure the first SD_freeException() does not do anything stupid */
	SD_I_setException(exception, SDStatus_SUCCESS, NULL);

	tmpvos = NULL;
	/* Let the user override the VO if the caller did not specify
	 * one explicitely */
	if (!vos || !vos->numNames)
		tmpvos = checkVOEnv();

	while ((plugin = nextPlugin(plugin)))
	{
		/* Free the exception set by the previous iteration */
		SD_freeException(exception);

		if (plugin->listServicesByHost)
		{
			result = plugin->listServicesByHost(type, host,
							    tmpvos ? tmpvos : vos,
							    exception);
		}
		if (result)
		{
			SET_OWNER(result, plugin);
			break;
		}
	}

	if (!plugin && exception && exception->status == SDStatus_SUCCESS) {
	    // filter the host and port from the result of
	    // listServices on the client side
	    result = SD_listServices(type, NULL, tmpvos ? tmpvos : vos, exception);
	    if(result != NULL && exception->status == SDStatus_SUCCESS) {
            // Check for the host name in the endpoint in this order:
            //    1. ^host$
            //    2. ^host:
            //    3. /host:
            //    4. /host/
            //    5. host
            // Run the filtering with the first matching pattern.
            // TODO: implement the filtering in bdii-c
            int f, fmatched;
            SD_matchEndpointHost matchFunction = NULL;
            for(f = 0; f < SD_num_matches; f++) {
                matchFunction = SD_matchEndpointHostFunctions[f];
                fmatched = 0;
                for(i = 0; i < result->numServices; i++) {
                    if(result->services[i]->endpoint != NULL) { 
                        if(matchFunction(result->services[i]->endpoint, host)) {
                            fmatched++;
                        }
                    }
                } // end for loop 
                SD_I_log4c_DEBUG("%d. filter matched %d times", f, fmatched);
                // break the loop, if there was a match and use that for filtering
                if(fmatched > 0) break;
            }
            // in the worst case, when there was no match,
            // matchFunction will point to the last function

            for(i = 0; i < result->numServices; i++) {
                if(result->services[i]->endpoint != NULL) { 
                    // check if the endpoint contains the host name
                    if(! matchFunction(result->services[i]->endpoint, host)) {
                        // no match. Delete the service entry and mark it as null 
                        // in the list
                        SD_freeService(result->services[i]);
                        result->services[i] = NULL;
                    }
                }
            } // end for loop 

            // Reorder the list so the nulls are at the end
            // Please Note: this algorithm may not be the best one but the 
            // number of expected elements should be quite limited, so it 
            // shouldn't be a real problem
            numServices = 0;
            for(i = 0; i < result->numServices; ++i) {
                if(NULL == result->services[i]){
                    // Get the next valid one
                    for(j = i+1; j < result->numServices; ++j){
                        if(NULL != result->services[j]){
                             // Swicth the two elements
                             result->services[i] = result->services[j];
                             result->services[j] = 0;
                             ++numServices;
                             break;
                        } 
                    }
                } else {
                    ++numServices;
                }
            } // end reordering loop
            result->numServices = numServices;
	    } // if result != NULL
	}

	if (tmpvos)
		SD_I_freeVOList(tmpvos);
	return result;
}

PUBLIC SDServiceList *SD_listServicesByData(const SDServiceDataList *data, const char *type, const char *site, const SDVOList *vos, SDException *exception)
{
	const SDPlugin *plugin = NULL;
	SDServiceList *result = NULL;
	SDVOList *tmpvos;

	/* Make sure the first SD_freeException() does not do anything stupid */
	SD_I_setException(exception, SDStatus_SUCCESS, NULL);

	if (!site)
    {
		site = getenv(SDEnv_SITE);
        /* treat empty string as if it does not exist */
        if (site && (strlen(site) == 0)) site = NULL;
    }
	tmpvos = NULL;
	/* Let the user override the VO if the caller did not specify
	 * one explicitely */
	if (!vos || !vos->numNames)
		tmpvos = checkVOEnv();

	while ((plugin = nextPlugin(plugin)))
	{
		/* Free the exception set by the previous iteration */
		SD_freeException(exception);

		result = plugin->listServicesByData(data, type, site,
			tmpvos ? tmpvos : vos,
			exception);
		if (result)
		{
			SET_OWNER(result, plugin);
			break;
		}
	}

	if (!plugin && exception && exception->status == SDStatus_SUCCESS)
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_NOPLUGIN);

	if (tmpvos)
		SD_I_freeVOList(tmpvos);
	return result;
}

PUBLIC void SD_freeServiceDataList(SDServiceDataList *serviceDataList)
{
	const SDPlugin *plugin;

	if (!serviceDataList)
		return;
	plugin = serviceDataList->_owner;

	if (plugin && plugin->freeServiceDataList)
		plugin->freeServiceDataList(serviceDataList);
	else
		SD_I_freeServiceDataList(serviceDataList);
}

PUBLIC void SD_freeService(SDService *service)
{
	const SDPlugin *plugin;

	if (!service)
		return;
	plugin = service->_owner;

	if (plugin && plugin->freeService)
		plugin->freeService(service);
	else
		SD_I_freeService(service);
}

PUBLIC void SD_freeServiceList(SDServiceList *serviceList)
{
	const SDPlugin *plugin;

	if (!serviceList)
		return;
	plugin = serviceList->_owner;

	if (plugin && plugin->freeServiceList)
		plugin->freeServiceList(serviceList);
	else
		SD_I_freeServiceList(serviceList);
}

PUBLIC void SD_freeServiceDetails(SDServiceDetails *serviceDetails)
{
	const SDPlugin *plugin;

	if (!serviceDetails)
		return;
	plugin = serviceDetails->_owner;

	if (plugin && plugin->freeServiceDetails)
		plugin->freeServiceDetails(serviceDetails);
	else
		SD_I_freeServiceDetails(serviceDetails);
}

PUBLIC void SD_freeException(SDException *exception)
{
	SD_I_freeException(exception);
}

/* PUBLIC FUNCTIONS OF THE INTERNAL API ***************************************/

PUBLIC int SD_registerPlugin(const SDPlugin *plugin, SDException *exception)
{
	const char *env;

	if (!plugin)
	{
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_NULLPLUGIN);
		return -1;
	}

	if (!plugin->name)
	{
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_NOPLUGINNAME);
		return -1;
	}

	if (getPluginByName(plugin->name))
	{
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_PLUGINEXISTS);
		return -1;
	}

	if (!plugin->getServiceDetails || !plugin->listAssociatedServices ||
			!plugin->listServicesByData)
	{
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_BADPLUGIN);
		return -1;
	}

	if (addPluginToList(&REGISTERED_PLUGINS, plugin))
	{
		SD_I_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY);
		return -1;
	}

	env = getenv(SDEnv_PLUGIN);
	if (!env)
	{
		if (addPluginToList(&LOADED_PLUGINS, plugin))
		{
			SD_I_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY);
			return -1;
		}
	}
	else
	{
		/* Recalculate the list and order of the plugins to use */
		char *plugList, *plugName, *tmp;

		plugList = dupstr(env);
		if (!plugList)
		{
			SD_I_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY);
			return -1;
		}

		/* Reset the list of useable plugins */
		LOADED_PLUGINS.numPlugins = 0;

		/* Split the plugin list by the commas and add the plugins to the useable
		 * list in the order they appear in the env. variable */
		tmp = plugList;
		while ((plugName = strtok(tmp, ",")))
		{
			tmp = NULL;
			plugin = getPluginByName(plugName);
			if (!plugin)
				continue;
			if (addPluginToList(&LOADED_PLUGINS, plugin))
			{
				SD_I_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY);
				free(plugList);
				return -1;
			}
		}
		free(plugList);
	}

	return 0;
}

PUBLIC void SD_unregisterPlugin(const SDPlugin *plugin)
{
	if (!plugin)
		return;

	delPluginFromList(&REGISTERED_PLUGINS, plugin);
	delPluginFromList(&LOADED_PLUGINS, plugin);
}


/* PRIVATE FUNCTIONS **********************************************************/

/* Convert the value of the SD_VO env. variable to an SDVOList */
PRIVATE SDVOList *checkVOEnv(void)
{
	SDVOList *volist;
	const char *vo;

	vo = getenv(SDEnv_VO);
	if (!vo)
		return NULL;

    if (strlen(vo) == 0)
        return NULL;

	volist = (SDVOList *) calloc(1, sizeof(SDVOList));
	if (!volist)
		return NULL;

	volist->names = (char **) malloc(sizeof(char *));
	if (!volist->names)
	{
		free(volist);
		return NULL;
	}

	volist->numNames = 1;
	volist->names[0] = dupstr(vo);
	if (!volist->names[0])
	{
		free(volist->names);
		free(volist);
		return NULL;
	}
	return volist;
}

PRIVATE int addPluginToList(SDPluginList *list, const SDPlugin *plugin)
{
	const SDPlugin **plugtmp;

	plugtmp = realloc(list->plugins, (list->numPlugins + 1) * sizeof(*plugtmp));
	if (!plugtmp)
		return -1;

	plugtmp[list->numPlugins++] = plugin;
	list->plugins = plugtmp;
	return 0;
}

PRIVATE void delPluginFromList(SDPluginList *list, const SDPlugin *plugin)
{
	const SDPlugin *old_plugin;
	int i;

	for (i = 0; i < list->numPlugins; i++)
	{
		if (strcmp(list->plugins[i]->name, plugin->name))
			continue;

		old_plugin = list->plugins[i];

		memmove(list->plugins + i, list->plugins + i + 1,
			(list->numPlugins - i - 1) * sizeof(*list->plugins));
		--list->numPlugins;

		/* Be nice to memory leak detectors */
		if (!list->numPlugins)
		{
			free(list->plugins);
			list->plugins = NULL;
		}
	}
}

/* Look for a plugin with the given name. */
PRIVATE const SDPlugin *getPluginByName(const char *pluginName)
{
	int i;

	for (i = 0; i < REGISTERED_PLUGINS.numPlugins; i++)
	{
		if (!strcmp(REGISTERED_PLUGINS.plugins[i]->name, pluginName))
			return REGISTERED_PLUGINS.plugins[i];
	}
	return NULL;
}

PRIVATE void tryToLoadPlugins(void)
{
	char *env, *plugList, *plugName, *tmp, *libName, *initFuncName;
	void (*initFunc)(void);
	void *handle;
	char **plugArray = NULL;
	int i, plugNum = 0;

	env = getenv(SDEnv_PLUGIN);
    /* fall-back to the default plugin, if  */
    /* the env. varibale is not set: "file" */
	if (!env)
		env = SD_DEFAULT_PLUGIN;

	plugList = dupstr(env);
	if (!plugList)
		return;

	/* If SDEnv_PLUGIN is defined, unload already loaded plugins */
	for (i = 0; i < REGISTERED_PLUGINS.numPlugins; i++)
		SD_unregisterPlugin(REGISTERED_PLUGINS.plugins[i]);

	tmp = plugList;

	plugName = strtok(tmp, ",");
	while (plugName)
	{
		plugArray = realloc(plugArray, (plugNum+1)*sizeof(char *));
		plugArray[plugNum++] = strdup(plugName);
		plugName = strtok(NULL, ",");
	}
	
	for (i = 0; i < plugNum; i++)
	{
		plugName = plugArray[i];
		libName = (char *) malloc(strlen(SDPlugin_Library_Prefix) +
			strlen(plugName) +
			strlen(SDPlugin_Library_Suffix) + 1);
		if (!libName)
			continue;

		strcpy(libName, SDPlugin_Library_Prefix);
		strcat(libName, plugName);
		strcat(libName, SDPlugin_Library_Suffix);

		handle = dlopen(libName, RTLD_LAZY);
		free(libName);
		if (!handle)
			continue;

		/* Look up the plugin initialization function */
		initFuncName = (char *) malloc(strlen(plugName) + strlen(SDPlugin_InitSymbol_Prefix) + 1);
		if (!initFuncName)
		{
			dlclose(handle);
			continue;
		}

		strcpy(initFuncName, SDPlugin_InitSymbol_Prefix);
		strcat(initFuncName, plugName);

		initFunc = dlsym(handle, initFuncName);
		free(initFuncName);
		if (!initFunc)
		{
			dlclose(handle);
			continue;
		}

		/* Call the plugin initialization function */
		initFunc();
	}

	for (i = 0; i < plugNum; i++)
	{
		free(plugArray[i]);
	}
	free(plugArray);
	free(plugList);

}

/* Return the next plugin to try. Passing NULL will return the first plugin.
 * If there are no more plugins, NULL is returned */
PRIVATE const SDPlugin *nextPlugin(const SDPlugin *plugin)
{
	int i;

	if (!LOADED_PLUGINS.numPlugins)
	{
		tryToLoadPlugins();
		if (!LOADED_PLUGINS.numPlugins)
			return NULL;
	}

	if (!plugin)
		return LOADED_PLUGINS.plugins[0];

	for (i = 0; i < LOADED_PLUGINS.numPlugins; i++)
		if (LOADED_PLUGINS.plugins[i] == plugin)
			break;

	if (i + 1 >= LOADED_PLUGINS.numPlugins)
		return NULL;

	return LOADED_PLUGINS.plugins[i + 1];
}

/* Portable replacement for strdup(). */

PRIVATE char *dupstr(const char *from)
{
	char *to;

	to = (char *) malloc(strlen(from) + 1);
	if (to != NULL) strcpy(to, from);
	return to;
}

/* End of file. */
