/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://eu-egee.org/partners/ for details on the copyright holders.
 * For license conditions see the license file or http://eu-egee.org/license.html
 *
 * Authors:
 *     Gabor Gombas <Gabor.Gombas@cern.ch>
 *     John Walk <J.A.Walk@rl.ac.uk>
 */

#ifndef SERVICE_DISCOVERY_I_H
#define SERVICE_DISCOVERY_I_H

/* This header file contains the private interface shared by all Service
   Discovery implementations. The functions defined in here are called by
   ServiceDiscovery.c only. The parameters are identical to the corresponding
   ones in ServiceDiscovery.h (q.v.). */

#include "ServiceDiscovery.h" /* For type definitions. */

/* Prefix of Service Discovery Plugin module names */
#define SDPlugin_Library_Prefix     "libglite-sd-c-"
/* Suffix of Service Discovery Plugin module names */
#define SDPlugin_Library_Suffix     ".so"
/* Prefix of the symbol name of the plugin's initialization function */
#define SDPlugin_InitSymbol_Prefix "SD_initPlugin_"
/* Prefix of the symbol name of the plugin's destructor function */
#define SDPlugin_DoneSymbol_Prefix "SD_donePlugin_"

/* GLOBAL TYPE DEFINITIONS  ***************************************************/

typedef struct
{
	/* Name of the plugin. */
	char *name;

	/* Mandatory plugin methods. */
	SDServiceDetails *(*getServiceDetails)(const char *serviceName, SDException *exception);
	SDServiceList *(*listAssociatedServices)(const char *serviceName, const char *type, const char *site, const SDVOList *vos, SDException *exception);
	SDServiceList *(*listServicesByData)(const SDServiceDataList *data, const char *type, const char *site, const SDVOList *vos, SDException *exception);

	/* Optional plugin methods. If they are not set, the core library
	 * will emulate them using the mandatory methods. */
	SDServiceList *(*listServices)(const char *type, const char *site, const SDVOList *vos, SDException *exception);
	SDService *(*getService)(const char *serviceName, SDException *exception);
	SDServiceDataList *(*getServiceData)(const char *serviceName, SDException *exception);
	char *(*getServiceDataItem)(const char *serviceName, const char *key, SDException *exception);
	char *(*getServiceSite)(const char *serviceName, SDException *exception);
	char *(*getServiceWSDL)(const char *serviceName, SDException *exception);
	SDServiceList *(*listServicesByHost)(const char *type, const char *host, const SDVOList *vos, SDException *exception);

	/* Optional plugin methods. If they are not set, a default implementation found in
	 * ServiceDiscoveryICommon.c will be used. If the plugin does not implement one or
	 * more of the previous group of optional methods then it should leave these
	 * methods as NULL. */
	void (*freeServiceDataList)(SDServiceDataList *serviceDataList);
	void (*freeService)(SDService *service);
	void (*freeServiceList)(SDServiceList *serviceList);
	void (*freeServiceDetails)(SDServiceDetails *serviceDetails);
	void (*freeVOList)(SDVOList *vos);
} SDPlugin;


/* PUBLIC FUNCTIONS ***********************************************************/

/**
 * The SD_registerPlugin function registers a Service Discovery plugin. The
 * name of the plugin will be used to decide which plugin to use at run time.
 *
 * @param  plugin             Descriptor of the plugin to register.
 * @param  exception          If not NULL, receives status of API call.
 *
 * @return 0 if the registration is successful or -1 if the API call fails.
 */
extern int SD_registerPlugin(const SDPlugin *plugin, SDException *exception);

/**
 * The SD_unregisterPlugin function unregisters a Service Discovery plugin that
 * was previously registered by SD_RegisterPlugin.
 *
 * @param  plugin             The plugin descriptor that was passed to SD_registerPlugin.
 */
extern void SD_unregisterPlugin(const SDPlugin *plugin);

/*
 * Common helper functions defined in ServiceDiscoveryI.c. Their prototype
 * matches the appropriate function in ServiceDiscovery.h
 */
extern void SD_I_setException(SDException *exception, int status, const char *reason);
extern void SD_I_freeServiceDataList(SDServiceDataList *serviceDataList);
extern void SD_I_freeService(SDService *service);
extern void SD_I_freeServiceList(SDServiceList *serviceList);
extern void SD_I_freeServiceDetails(SDServiceDetails *serviceDetails);
extern void SD_I_freeException(SDException *exception);
extern void SD_I_freeVOList(SDVOList *vos);



/* Name of the environment variable holding the log level. */
#define SDEnv_LOG_LEVEL      "GLITE_SD_LOG_LEVEL"
/* Name of the environment variable holding the log filename. */
#define SDEnv_LOG_FILE       "GLITE_SD_LOG_FILE"

/** 
 * The SD_I_log4c_init() function initializes the common logging 
 * facility for service discovery plugins.
 *
 * GLITE_SD_LOG_LEVEL environmental variable sets the log level,
 * if left empty, then no logging is performed. Acceptable values
 * are: DEBUG, INFO, WARN, ERROR
 *
 * GLITE_SD_LOG_FILE may specify a file, where the logs are written.
 * If not set, then they are written to stdout.
 */
extern void SD_I_log4c_init();
/**
 * The SD_I_log4c_done() function destroys the context of the
 * common logging facility, closing open file handles, etc.
 */
extern void SD_I_log4c_done();

/*
 * Log levels for the SD_I_log4c_printf() function, however
 * it is recommended to use the provided macros!
 */
typedef enum {
    SD_I_LOG4C_DEBUG,
    SD_I_LOG4C_INFO,
    SD_I_LOG4C_WARN,
    SD_I_LOG4C_ERROR,
    SD_I_LOG4C_NONE,
    SD_I_LOG4C_UNKNOWN
} SD_I_log4c_LogLevel;

extern void SD_I_log4c_printf(SD_I_log4c_LogLevel loglevel, 
    const char *file, const char *function, const int line,
    const char *format, ...);

#define SD_I_log4c_DEBUG(fmt, ...) SD_I_log4c_printf(SD_I_LOG4C_DEBUG, __FILE__, __func__, __LINE__, fmt,##__VA_ARGS__)
#define SD_I_log4c_INFO(fmt, ...) SD_I_log4c_printf(SD_I_LOG4C_INFO, __FILE__, __func__, __LINE__, fmt,##__VA_ARGS__)
#define SD_I_log4c_WARN(fmt, ...) SD_I_log4c_printf(SD_I_LOG4C_WARN, __FILE__, __func__, __LINE__, fmt,##__VA_ARGS__)
#define SD_I_log4c_ERROR(fmt, ...) SD_I_log4c_printf(SD_I_LOG4C_ERROR, __FILE__, __func__, __LINE__, fmt,##__VA_ARGS__)

#endif /* SERVICE_DISCOVERY_I_H */

/* End of file. */
