/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://eu-egee.org/partners/ for details on the copyright holders.
 * For license conditions see the license file or http://eu-egee.org/license.html
 *
 * Authors:
 *     Gabor Gombas <Gabor.Gombas@cern.ch>
 *     John Walk <J.A.Walk@rl.ac.uk>
 */

/* This file provides the implementation for the housekeeping functions for
   the ServiceDiscoveryAPI (i.e. part of the interface defined in the file
   ServiceDiscoveryI.h). */

#include <stdio.h>                  /* for NULL */
#include <stdlib.h>                 /* for free() */
#include <string.h>                 /* for strcmpy() and strlen() */

#include "ServiceDiscoveryI.h"      /* Private interface to implementation. */

#define PRIVATE static
#define PUBLIC /* empty */

static char *SD_EMPTY_MESSAGE = "";		/* default error message */
static char *SD_NOMEM_MESSAGE = "Out of memory";/* error message when there were not enough
						   memory for even setting the message */
                                                                                
/* PUBLIC FUNCTIONS ***********************************************************/


/**
  * Function SD_I_setException initialises an existing SDException structure with
  * the given status and message. Ensures that "reason" is NULL if and only if
  * "status" is SDStatus_SUCCESS, in accordance with the API.
  *
  * @param  exception        Exception to initialise.
  * @param  status           Status value (see ServiceDiscovery.h).
  * @param  reason           Reason for failure (may be NULL).
  *
  * @return Nothing.
  */

PUBLIC void SD_I_setException(SDException *exception, int status, const char *reason)
{
	if (!exception)
	{
		return;
	}

	exception->status = status;
	if (status == SDStatus_SUCCESS)
	{
		exception->reason = NULL;
	}
	else
	{
		if (reason != NULL)
		{
			exception->reason = (char *) malloc(strlen(reason) + 1);
			if (exception->reason != NULL)
			{
				strcpy(exception->reason, reason);
			}
			else
			{
				exception->reason = SD_NOMEM_MESSAGE;
			}
		}
		else
		{
			exception->reason = SD_EMPTY_MESSAGE;
		}
	}
}


/**
 * Default implementation of the freeServiceDataList method.
 *
 * @param  A SD_freeServiceDataList.
 */

PUBLIC void SD_I_freeServiceDataList(SDServiceDataList *serviceDataList)
{
	int i;

	if (serviceDataList)
	{
		if (serviceDataList->items)
		{
			for (i = 0; i < serviceDataList->numItems; ++i)
			{
				if (serviceDataList->items[i].key)
				{
					free(serviceDataList->items[i].key);
				}
				if (serviceDataList->items[i].value)
				{
					free(serviceDataList->items[i].value);
				}
			}
			free(serviceDataList->items);
		}
		free(serviceDataList);
	}
}


/**
 * Default implementation of the freeService method.
 *
 * @param  A SD_freeService.
 */
                                                                                
PUBLIC void SD_I_freeService(SDService *service)
{
	if (service)
	{
		if (service->name) free(service->name);
		if (service->type) free(service->type);
		if (service->endpoint) free(service->endpoint);
		if (service->version) free(service->version);
		free(service);
	}
}


/**
 * Default implementation of the freeServiceList method.
 *
 * @param  As SD_freeServiceList.
 */

PUBLIC void SD_I_freeServiceList(SDServiceList *serviceList)
{
	int i;

	if (serviceList)
	{
		if (serviceList->services)
		{
			for (i = 0; i < serviceList->numServices; ++i)
			{
				if (serviceList->services[i])
				{
					SD_I_freeService(serviceList->services[i]);
				}
			}
			free(serviceList->services);			
		}
		free(serviceList);			
	}
}


/**
 * Default implementation of the freeServiceDetails method.
 *
 * @param  As SD_freeServiceDetails.
 */

PUBLIC void SD_I_freeServiceDetails(SDServiceDetails *serviceDetails)
{
	if (serviceDetails)
	{
		if (serviceDetails->name) free(serviceDetails->name);
		if (serviceDetails->type) free(serviceDetails->type);
		if (serviceDetails->endpoint) free(serviceDetails->endpoint);
		if (serviceDetails->version) free(serviceDetails->version);
		if (serviceDetails->site) free(serviceDetails->site);
		if (serviceDetails->wsdl) free(serviceDetails->wsdl);
		if (serviceDetails->administration) free(serviceDetails->administration);
		if (serviceDetails->vos) SD_I_freeVOList(serviceDetails->vos);
		if (serviceDetails->associatedServices) SD_I_freeServiceList(serviceDetails->associatedServices);
		if (serviceDetails->data) SD_I_freeServiceDataList(serviceDetails->data);
		free(serviceDetails);
	}
}


/**
 * Implementation of SD_freeException.
 *
 */
PUBLIC void SD_I_freeException(SDException *exception)
{
	if (exception)
	{
		if (exception->reason &&
		    exception->reason != SD_EMPTY_MESSAGE &&
		    exception->reason != SD_NOMEM_MESSAGE)
		{
			free(exception->reason);
			exception->reason = NULL;
		}
		/* don't free exception itself, it's not on the heap */
	}
}


/**
 * Frees an SDVOList - there's no public interface to this in the API because
 * we never return a standalone VO list to the caller.
 *
 * @param  vos        VO list to free.
 *
 * @param  Nothing.
 */

PUBLIC void SD_I_freeVOList(SDVOList *vos)
{
	int i;

	if (vos)
	{
		if (vos->names)
		{
			for (i = 0; i < vos->numNames; ++i)
			{
				if (vos->names[i])
				{
					free(vos->names[i]);
				}
			}
			free(vos->names);
		}
		free(vos);
	}
}

/* End of file. */
